"""Read files with data hidden inside them through Steganography
Supported files:
    Standard:
        png
        gif
    If you install PIL:
        jpg
        bmp
        (most other image formats)

Functions:

    readImage(image_file: str, output_file: str, depth: int) -> (bytes) data

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["readImage"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

from .SimpleImageReader import readImageSimple

advanced = True
try:
    from .AdvancedImageReader import readImageAdvanced
except ModuleNotFoundError:
    advanced = False


def readImage(image_file: str, output_file: str, depth: int) -> bytes:
    """Read steganography data in almost any image file (If PIL is installed)

        Parameters:
            image_file (str):  The path to the image to read
            output_file (str): The path to the file where the output data should be stored
            depth (int):       The bit depth to read

        Returns:
            byte_string (bytes): Hidden data
    """
    if advanced:
        return readImageAdvanced(image_file, output_file, depth)
    else:
        return readImageSimple(image_file, output_file, depth)



"""Read files with data hidden inside them through Steganography
Supported files:
    png
    gif

Functions:

    readImageSimple(image_file: str, output_file: str, depth: int) -> (bytes) data

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["readImageSimple"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

import tkinter
from .RawReader import readRaw

img = None
tk = tkinter.Tk()
tk.withdraw()


def readImageSimple(image_file: str, output_file: str, depth: int) -> bytes:
    """Read steganography data in png and gif images

        Parameters:
            image_file (str):  The path to the image to read
            output_file (str): The path to the file where the output data should be stored
            depth (int):       The bit depth to read

        Returns:
            byte_string (bytes): Hidden data
    """
    global img, tk
    img = tkinter.PhotoImage(file=image_file, master=tk)
    tk.image = img
    pixels = []
    for index in range(img.width() * img.height()):
        pixels.append(img.get(index % img.width(), index // img.width()))
    pixels = [channel for pixel in pixels for channel in pixel]

    byte_string = readRaw(pixels, depth)

    try:
        with open(output_file, "wb") as f:
            f.write(byte_string)
    except TypeError:
        pass

    return byte_string



"""Create and read files with data hidden inside them through Steganography
Supported files:
    Standard:
        wav
        png
        gif
    If you install PIL
        jpg
        bmp
        (most other image formats)

Functions:

    read.AudioReader.readAudio(audio_file: str, output_file: str, depth: int) -> (bytes) data
    read.ImageReader.readImage(image_file: str, output_file: str, depth: int) -> (bytes) data
    read.RawReader.readRaw(initial_data: Union[array.array, list], depth: int) -> (bytes) data
    write.AudioWriter.writeAudio(in_audio_file: str, out_audio_file: str, data_file: str, depth: int) ->
                                                                                            (list) Audio data as a list
    write.ImageWriter.writeImage(in_image_file: str, out_image_file: str, data_file: str, depth: int) ->
                                                                                            (list) Image data as a list
    write.RawWriter.writeRaw(initial_data: array.array, file_data_bits: Union[array.array, list], depth: int) ->
                                                                                            (array.array) Raw byte data


Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["AudioReader", "ImageReader", "RawReader", "AudioWriter", "ImageWriter", "RawWriter"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

from .read import AudioReader, ImageReader, RawReader
from .write import AudioWriter, ImageWriter, RawWriter



"""Read files with data hidden inside them through Steganography

Functions:

    readRaw(initial_data: Union[array.array, list], depth: int) -> (bytes) data

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["readRaw"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

import array
import math
from typing import Union


def readRaw(initial_data: Union[array.array, list], depth: int) -> bytes:
    """Read steganography data in raw byte data

        Parameters:
            initial_data (Union[array.array, list]): The path to the image to read
            depth (int):                             The bit depth to read

        Returns:
            byte_string (bytes): Hidden data
    """
    bit_file_data = []
    byteslist = []

    bit_depth = int(math.log2(depth))
    dec_remainder = 2 ** bit_depth

    for index in range(0, len(initial_data)):
        data = format(initial_data[index] % dec_remainder, "b")
        data = "0" * (bit_depth - len(data)) + data
        for bit in data:
            bit_file_data.append(int(bit))
    for index in range(0, len(bit_file_data), 8):
        byte = ""
        for bit in bit_file_data[index:index+8]:
            byte += str(bit)
        byteslist.append(int(byte, 2))
    return bytes(byteslist)



"""Read files with data hidden inside them through Steganography
Supported files:
    wav

Functions:

    readAudio(audio_file: str, output_file: str, depth: int) -> (bytes) data

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["readAudio"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

import wave
import array
from .RawReader import readRaw


def readAudio(audio_file: str, output_file: str, depth: int) -> bytes:
    """Read steganography data in a wav file

        Parameters:
            audio_file (str):  The path to the audio file to read
            output_file (str): The path to the file where the output data should be stored
            depth (int):       The bit depth to read

        Returns:
            byte_string (bytes): Hidden data
    """
    with wave.open(audio_file, 'rb') as out_wavefile:
        sizes = {1: 'B', 2: 'h', 4: 'i'}
        samp_size = out_wavefile.getsampwidth()
        frmt = sizes[samp_size]

        initial_data = array.array(frmt)
        initial_data.frombytes(out_wavefile.readframes(out_wavefile.getnframes()))

    byte_string = readRaw(initial_data, depth)

    try:
        with open(output_file, "wb") as f:
            f.write(byte_string)
    except TypeError:
        pass

    return byte_string



"""Create files with data hidden inside them through Steganography

Functions:

    writeRaw(initial_data: array.array, file_data_bits: Union[array.array, list], depth: int) ->
                                                                                             (array.array) Raw byte data

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["writeRaw"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

import math
import array
from typing import Union


def writeRaw(initial_data: array.array, file_data_bits: Union[array.array, list], depth: int) -> array.array:
    """Write steganography data in raw byte data

        Parameters:
            initial_data (array.array):                The array of data to overwrite
            file_data_bits (Union[array.array, list]): The array to overwrite with
            depth (int):                               The bit depth to write

        Returns:
            final_data (array.array): Raw byte data as an array.array
    """
    final_data = array.array(initial_data.typecode)
    bit_depth = int(math.log2(depth))
    dec_remainder = 2 ** bit_depth

    for index in range(0, len(initial_data)):
        addition = ""
        for bit in file_data_bits[index * bit_depth:(index + 1) * bit_depth]:
            addition += str(bit)
        if addition == "":
            addition = "0"
        final_data.append(initial_data[index] // dec_remainder * dec_remainder + int(addition, 2))
    return final_data



"""Create and read files with data hidden inside them through Steganography
Supported files:
    Standard:
        wav
        png
        gif
    If you install PIL
        jpg
        bmp
        (most other image formats)

Functions:

    cli() -> (None)

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["cli"]
__author__ = "Alexander Bisland"
__version__ = "1.1.2"
supported_from = "3.8.1"

from StegaSaurus.read import AudioReader, ImageReader
from StegaSaurus.write import AudioWriter, ImageWriter
import argparse
import os


def cli() -> None:
    """Create and read files with data hidden inside them through Steganography

        Parameters:
            (None)

        Returns:
            Nothing (None): Null
    """
    parser = argparse.ArgumentParser(description='(Optionally) Dependency-free library for steganography')
    parser.add_argument('input', type=argparse.FileType('r'), help='Input file path')
    parser.add_argument('--read', '-r', dest='option', action='store_const',
                        const="r", default="w", help='Decrypt Seganography file')
    parser.add_argument('--write', '-w', dest='option', action='store_const',
                        const="w", help='Encrypt Seganography file')
    parser.add_argument('--audio', '-a', dest='type', action='store_const',
                        const="a", default="i", help='Perform operation on audio file')
    parser.add_argument('--image', '-i', dest='type', action='store_const',
                        const="i", help='Perform operation on image file')
    parser.add_argument('--no-output', dest='no_output', action='store_const',
                        const=True, default=False, help='No output file for decryption')
    parser.add_argument('--output', type=argparse.FileType('w'), default=None, help='Output file path')
    parser.add_argument('--data', type=argparse.FileType('r'), default=None, help='Data file path (write only)')
    parser.add_argument('--depth', '-d', type=int, default=2, help='Bit depth (default is 2)')
    parser.add_argument('--raw', type=str, default=None, help='Raw data for if data file is not used')

    args = parser.parse_args()

    data = os.path.abspath(args.data.name)
    if args.raw is not None:
        data = args.raw
    if args.no_output:
        output_file = None
    else:
        output_file = os.path.abspath(args.output.name)
    if args.option == "w":
        if args.output is None:
            print("Output file is required")
            exit()
        if data is None:
            print("Data file is required")
            exit()
        if args.type == "a":
            AudioWriter.writeAudio(os.path.abspath(args.input.name), os.path.abspath(args.output.name), data,
                                   args.depth)
        else:
            ImageWriter.writeImage(os.path.abspath(args.input.name), os.path.abspath(args.output.name), data,
                                   args.depth)
    else:
        if args.output is None:
            print("Output file is required")
            exit()
        if args.type == "a":
            resp = AudioReader.readAudio(os.path.abspath(args.input.name), output_file, args.depth)
            if output_file is None:
                print(resp)
        else:
            resp = ImageReader.readImage(os.path.abspath(args.input.name), output_file, args.depth)
            if output_file is None:
                print(resp)


if __name__ == "__main__":
    cli()



"""Read files with data hidden inside them through Steganography
Supported files:
    Standard:
        wav
        png
        gif
    If you install PIL:
        jpg
        bmp
        (most other image formats)

Functions:

    AudioReader.readAudio(audio_file: str, output_file: str, depth: int) -> (bytes) data
    ImageReader.readImage(image_file: str, output_file: str, depth: int) -> (bytes) data
    AdvancedImageReader.readImageAdvanced(image_file: str, output_file: str, depth: int) -> (bytes) data
    SimpleImageReader.readImageSimple(image_file: str, output_file: str, depth: int) -> (bytes) data
    RawReader.readRaw(initial_data: Union[array.array, list], depth: int) -> (bytes) data

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["AudioReader", "ImageReader", "RawReader", "AdvancedImageReader", "SimpleImageReader"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

from StegaSaurus.read import ImageReader
from StegaSaurus.read import AdvancedImageReader
from StegaSaurus.read import SimpleImageReader
from StegaSaurus.read import AudioReader
from StegaSaurus.read import RawReader



"""Create files with data hidden inside them through Steganography
Supported files:
    Standard:
        wav
        png
        gif
    If you install PIL:
        jpg
        bmp
        (most other image formats)

Functions:

    writeImage(in_image_file: str, out_image_file: str, data_file: str, depth: int) -> (list) Image data as a list

Misc variables:

    __all__
    __author__
    __version__
    supported_from

Other Info:
    Steganography is the technique of hiding secret data within an ordinary, non-secret, file or message in order to
    avoid detection; the secret data is then extracted at its destination. The use of steganography can be combined with
    encryption as an extra step for hiding or protecting data.
"""

__all__ = ["writeImageAdvanced"]
__author__ = "Alexander Bisland"
__version__ = "1.0.2"
supported_from = "3.8.1"

from .SimpleImageWriter import writeImageSimple

advanced = True
try:
    from .AdvancedImageWriter import writeImageAdvanced
except ModuleNotFoundError:
    advanced = False


def writeImage(in_image_file: str, out_image_file: str, data_file: str, depth: int) -> list:
    """Write steganography data to almost any image file (If PIL is installed)

        Parameters:
            in_image_file (str):  The path to the image to write
            out_image_file (str): The path to the file where the output data should be stored
            data_file (str):      The path to the data file to hide
            depth (int):          The bit depth to write

        Returns:
            final_data (list): Image data as a list
    """
    if advanced:
        return writeImageAdvanced(in_image_file, out_image_file, data_file, depth)
    else:
        return writeImageSimple(in_image_file, out_image_file, data_file, depth)
